
#include "Steiner.h"


struct NotAMutex {
    void lock() {};
    void unlock() {};
};

// global variables for the parallel algorithm:
//   global steiner tree

//Steiner globalX;
NotAMutex steinerMapMutex;


//   best solution placeholder, and its mutex for writing in
HyperSteinerForest bestForest;
std::mutex bestSolutionMutex;
std::mutex printMutex;

int globalCount;
int steinerKeys;
int steinerReused;
vector<int> threadProgress;
std::mutex globalCountMutex;
vector<int> steinerNodes;

/*! Get subset:
 * get elements from 'original' at indices given in subset
 */
Vertex_Set getSubset(const vector<int> &original, const  vector<int> &subset)
{
    Vertex_Set S;
    int i = subset.size();
    while (i-- > 0) {
        S.insert(original[subset[i]]);
    }
    return S;
}

Vertex_Set getSubsubset(const vector<int>& original, const vector<int> &subset, const vector<int> &subsubset)
{
    Vertex_Set S;
    int i = subsubset.size();
    while (i-- > 0) {
        S.insert(original[subset[subsubset[i]]]);
    }
    return S;
}

void printVertexSet(const Vertex_Set &S)
{
    for (auto x : S) {
        std::cout << x << ' ';

    }
}

/* computes the complement of set Ta (as a vector) in {0,1,... (max-1)}
 */
vector<int> complement(const vector<int>& Ta, int max)
{
    unsigned int ja = 0;
    vector<int> Tb;
    for (int i = 0; i < max; i++) {
        if (ja >= Ta.size() || Ta[ja] != i) {
            Tb.push_back(i);
        }
        else {
            ja++;
        }
    }
    return Tb;
}


/* enumerate subsets of {0,1,... ,(totalSize-1)}
 * Starting with S (size SLength)
 * Until last subset of size maxSubsetSize
 *
 * usage:
 *  int totalSize =  ...//size of list to enumerate
 *  int maxSubsetSize =  ... //maximum size of a subset we're interested in
 *  vector<int> S; //array containing indices of successive subsets
 *  S.push_back(0); //to consider only non-empty arrays (remove this line to keep empty set)
 *  do  {
 *    ...
 *  } while (nextSubstet(S, totalSize, maxSubsetSize));
 *
 * NB: elements of S are sorted and must remain this way.
 */
bool nextSubset(vector<int> &S, int totalSize, int maxSubsetSize)
{
    //sanity check: do not allow subsets bigger than total size
    if (maxSubsetSize > totalSize) maxSubsetSize = totalSize;

    //single vairable for size of S
    int SLength = S.size();

    // first step: find the position that needs to be incremented
    // most frequent: right most-position
    int i = SLength - 1;
    // it can be incremented only if it is strictly smaller than m (which is the max value allowed)
    int m = totalSize - 1;

    while (i >= 0 && S[i] == m) {	// while S[i] is already at the max. value
        i--;			//go to the left
        m--;			//the new max value is 1 lower
    }

    // second step: check if we reached the last subset of the current size
    if (i < 0) {
        // no value can be incremented: we need to increase the size
        SLength++;
        S.push_back(0);
        if (SLength > maxSubsetSize) {
            // we're already at the final size: enumeration is complete
            return false;
        }
        //start again with a 0 in first position (other positions will be filled up afterwards)
        //for the very first set: SLength is 0, so we can't update S[i]
        i = 0;
        if (SLength > 0) {
            S[i] = 0;
        }
    }
    else {
        // position i can be incremented, so we do that
        S[i]++;
    }
    // third step: position i has been incremented, 
    // positions to the right of i must be given the smallest possible values: S[i]+1, S[i]+2, etc.
    while (i < SLength - 1) {
        i++;
        S[i] = S[i - 1] + 1;
    }
    return true;
}


bool nextKSubset(vector<int> &S, int totalSize, int maxSubsetSize, int k)
{
    //sanity check: do not allow subsets bigger than total size
    if (maxSubsetSize > totalSize) maxSubsetSize = totalSize;

    //single vairable for size of S
    int SLength = S.size();

    if (SLength>0 && S[SLength - 1] + k < totalSize) {
        S[SLength - 1] += k;
    }
    else {
        for (int i = 0; i < k; i++) {
            if (!nextSubset(S, totalSize, maxSubsetSize)) {
                return false;
            }
        }
    }
    return true;
}

/*
 * permutes the elements of an array S where all values are different,
 * so that successive calls to nextPermutation visit all possible permutations of the array.
 * Returns false if we've reached a sorted permutation (so the very first S must be sorted for using in a while loop)
 * min must be the minimal value of S.
 *
*/
bool nextPermutation(vector<int> &S, int min)
{
    if (S.size() < 2)
        return false;
    //using an hamiltonian path in the permutation graph
    //following only "pancake" flips, and minimizing the total size of the flips.

    //See Essed & Therese, The Harassed Waitress Problem, FUN 2014

    //first step : find the flip length
    // = longest decreasing prefix, up to circular permutation
    //(example: in 8 5 4 10 9 6 3, the flip covers the "almost" decreasing part 8 5 4 10 9)

    unsigned int flipEnd = 0;
    int inc = 0;
    while (flipEnd < S.size() - 1) {
        if (S[flipEnd] < S[flipEnd + 1])
            inc++;
        if (inc == 2 || (inc == 1 && S[flipEnd + 1] < S[0]))
            break;
        flipEnd++;
    }

    //second step: perform the flip
    int a = 0;
    int b = flipEnd;
    while (b > a) {
        int t = S[a];
        S[a] = S[b];
        S[b] = t;
        b--;
        a++;
    }
    //if we've just flipped the whole sequence, with S[0] minimal, then we've reached the sorted sequence.
    if (flipEnd == S.size() - 1 && S[0] == min)
        return false;
    return true;
}

/*
 * Prints an array of integer, under the following format:
 * length\t S[0] S[1] ... S[length-1]\n
 */
void print(const vector<int> &S)
{
    bool coma = false;
    for (int i : S) {
        if (coma) std::cout << ",";
        std::cout << i;
        coma = true;
    }
    std::cout << "\n";
}


/* print all sets present in a Steiner map, and the best value accessible for each.
* sets of size 1 or those which are not accessible from a single source are not printed.
*/
void printQuickSteiner(Steiner& X) {
    for (auto &i : X) {
        //i has the following structur: ( (set, source) , value)
        Key k = i.first;
        if (k.second != -1 || i.second.value == INF || k.first.size() < 2) continue;
        std::cout << " Cover [ ";
        printVertexSet(k.first);
        std::cout << "] \t  opt. value = " << i.second.value<<endl;
    }
}
/* looks up the steiner map for a given source and cover-set
* if the key is not present, we assume it is unreachable (+INF)
*/

void printSteinerStats(Steiner& X, int id) {
    int keys = 0;
    int nonsingletonkeys = 0; 
    int finitenonsingletonSets = 0;
    int sets = 0;
    int singletons = 0;
    int finiteSets = 0;
    for (auto &i : X) {
        keys++;
        Key k = i.first;
        if (k.first.size() > 1) nonsingletonkeys++;
        if (k.second != -1) continue;
        sets++;
        if (k.first.size() < 2) singletons++;
        if (i.second.value != INF) finiteSets++;
        if (i.second.value != INF&& k.first.size()>1) finitenonsingletonSets++;
    }
    std::cout << "*** Steiner Map statistics for Thread "<<id<<" ***\n";
    std::cout << "  "<<keys<<" keys covering "<< sets<<" sets, of which "<<finiteSets<<" are reachable.\n";
    std::cout << " Singletons : "<<(keys-nonsingletonkeys)<<" keys for "<<singletons<<" sets \n";
    std::cout << " Non-singletons : " << (nonsingletonkeys) << " keys for " << (sets-singletons) << " sets, of which " << finitenonsingletonSets << " are reachable. \n";
    std::cout << "************************************************\n";

}

Weight getSteinerValue(Steiner &X, Vertex_Set &T, int& source)
{

    steinerMapMutex.lock();
  
    //check that steiner trees over T have indeed been computed
    if (g_check_steiner_tree) assert(X.count(Key(T, -1)) == 1);
    //look for  the right key
    Steiner::iterator i = X.find(Key(T, source));
    bool found = i != X.end();
    steinerMapMutex.unlock();
    if (found) {
        //key is here: return the value
        return i->second.value;
    }
    //key is not here, not reachable
    return INF;
}


vector<int> enumerateSteinerNodes(const vector<int>& S, const vector<int>& T, int n, const vector<HyperArc>& hyperArcs, numeric::ublas::matrix<double> sp) {
    vector<int> sNodes;
    int nS = 0;
    int nnS = 0;
    vector<int> FullT = T;
    vector<int> FullS = S;
    for (const HyperArc& h : hyperArcs) {
        FullT.insert(FullT.end(), h.substrates.begin(), h.substrates.end());
        FullS.insert(FullS.end(), h.products.begin(), h.products.end());
    }
    for (int u = 0; u < n; u++) {
        bool steiner = true;
        if (std::find(FullS.begin(), FullS.end(), u) != FullS.end()) {
            steiner = true;
        }
        else {
            for (int v = 0; v < n; v++) {
                double suv = sp(u, v);
                if (!(suv > 0)) continue;
                bool steinerWithV = false;
                for (int t : FullT) {
                    if (suv + sp(v, t) != sp(u, t)) {
                        steinerWithV = true;
                        break;
                    }
                }
                if (!steinerWithV) {
                    steiner = false;
                    break;
                }
            }
        }
        if (steiner) {
            nS++;
            sNodes.push_back(u);
        }
        else {
            nnS++;
        }
    }
    std::cout << "Steiner nodes : " << nS << "    (non-steiner:" << nnS <<")   ";
    if (g_verbose > 1) {
        print(sNodes);
    }
    else {
        std::cout << endl;
    }
    return sNodes;
}

void checkKeyIsPresent(const Steiner &X, const Vertex_Set &T, const string& errorMsg) {
    if (g_check_steiner_tree) {
        steinerMapMutex.lock();
        if (X.count(Key(T, -1)) != 1) {
            std::cout << "error: missing key. count=" << X.count(Key(T, -1)) << ", T = ";
            printVertexSet(T);
            std::cout << "\t\t" << errorMsg << endl;
        }
        steinerMapMutex.unlock();
    }
}

bool steinerHasBeenDone(Steiner &X, Vertex_Set& T) {
    steinerMapMutex.lock();
    bool r = X.count(Key(T, -1)) == 1;
    steinerMapMutex.unlock();
    return r;
}



/* Compute all steiner trees for any subset of T and any vertex i
 *   X: steiner map containing possibly the solutions for subsets of T (no need to compute them again)
 *      results will be stored here:
 *      for any pair<Vertex_Set, source>, give optimal tree weight to cover the vertex set from source. No key added for INF
 *      also, for <Vertex_Set, -1>: add an entry for each set which have been visited (with value being the optimal, which may be INF)
 *   T: list of vertex indices (0<= <nVertex)
 *   nVertex: size of the graph
 *   maxTargetSubset: consider only small subsets of T. use non-positive value to disable this and visit all T
 *   sp : shortest path matrix (PCC(i,j) gives shortest path from i to j, infinity() otherwise
 *
 * Out: update the steiner Map X; no return otherwise
 */
void updateSteiner(Steiner &X, const vector<int> &T, const int nVertex,  int maxTargetSubset, const DistanceMatrix& sp)
{

    //check parameter maxTargetSubset
    if (maxTargetSubset <= 0) {
        maxTargetSubset = T.size();
    }

    //profiling variables: evaluate memory usage and gain of speed.
    int countKeys = 0;
    int countReachable = 0;
    int countSets = 0;

    //print debugging information
    bool debugSteiner = g_verbose > 2;

    //single variable for length of T
    //const int TLength = T.size();

    //successive subsets of T, start with singleton
    vector<int> Tp;
    Tp.push_back(0);

    if (debugSteiner) cout << "------\t Steiner update. Former size  : " << X.size() << endl;

    // iterate over Tp \subset T
    do {
        //get Tp in "Vertex_Set" format (using std::set instead of a list)
        Vertex_Set TpSet = getSubset(T, Tp);

        if (debugSteiner) {
            cout << "------\t\t[";
            printVertexSet(TpSet);
            cout << "]\t";
        }

        //if already done: skip to next subset.


        if (steinerHasBeenDone(X,TpSet)) {
            if (debugSteiner) cout << "Already done\n";
            if (g_keep_stats) {
                globalCountMutex.lock();
                steinerReused++;
                globalCountMutex.unlock();
            }
            continue;//next subset Tp
        }
        if (g_keep_stats) {
            globalCountMutex.lock();
            steinerKeys++;
            globalCountMutex.unlock();
        }


        //now TpSet has to be computed
        countSets++;

        //placeholder for  best way to cover TpSet
        SteinerValue veryBest;
        veryBest.value = INF;

        //enumerate  source vertices
        //for (int i = 0; i < nVertex; i++) {
        for (int i: steinerNodes) {
            //this source yields Key(TpSet, i)
            countKeys++;
            //placeholder for best tree from this source to cover Tp
            SteinerValue best;
            best.value = INF;

            //test size of Tp:
            if (Tp.size() == 1) {
                //base case: |Tp|=1; best solution consists of a shortest path
                best.singlePath = true;
                best.value = SP(i, T[Tp[0]]);
                if (best.value < INF)
                    countReachable++;
            }
            else {
                //general case: |Tp|>=2
                //check if there is a path from i to every t in T 
                bool reachable = true;
                for (int t : Tp) {
                    if (debugSteiner && reachable && SP(i, T[t]) == INF) cout << "!" << i<<"("<< T[t] <<")";
                    reachable = reachable && (SP(i, T[t]) < INF);
                }
                //if  some vertex is not reachable, no need to keep going
                if (!reachable) continue; //next source vertex i
                if (debugSteiner) {
                    cout << "$" << i;
                }
                //now Tp is reachable from the source, we need to compute the best steiner tree.
                countReachable++;
                best.singlePath = false;

                //subsubset of T
                vector<int> Tpa;
                Tpa.push_back(0);
                //Iterate Tpa over subsets of Tp (with |Tpa|<=|Tp|/2 )
                do {
                    //Compute Tpb = Tp\Tpa
                    vector<int> Tpb = complement(Tpa, Tp.size());

                    //Get Tpa and Tpb in "vertex set" format
                    Vertex_Set TpaSet = getSubsubset(T, Tp, Tpa);
                    Vertex_Set TpbSet = getSubsubset(T, Tp, Tpb);

                    //enumerate all steiner nodes 
                    for (int j : steinerNodes) {
                    //for (int j = 0; j < nVertex; j++) {
                        //distance to get from the source to steiner node
                        Weight wij = SP(i, j);

                        if (wij == INF) continue; //next steiner node j

                        //we have a valid steiner node here

                        checkKeyIsPresent(X, TpaSet, "Steiner - Tpa");
                        checkKeyIsPresent(X, TpbSet, "Steiner - Tpb");
                        Weight xa = getSteinerValue(X, TpaSet, j);
                        Weight xb = getSteinerValue(X, TpbSet, j);

                        //check that both subtrees are feasible
                        if (xa == INF || xb == INF) continue;//next steiner node j

                        //value of this partition
                        Weight x = xa + xb + wij;
                        if (x >= best.value) continue;//next steiner node j

                        //this partition is better than current best: remember it
                        best.value = x;
                        best.steinerVertex = j;
                        best.leftPartition = TpaSet;
                        best.rightPartition = TpbSet;
                    }  //for j
                } while (nextSubset(Tpa, Tp.size(), (int)Tp.size() / 2));	//enumerate Tpa
            }			//if [Tp|==1
            //Now best is computed in both cases (defaults to infinity if no solution)

            //record only if not INF
            if (best.value < INF) {
                steinerMapMutex.lock();
                X[Key(TpSet, i)] = best;
                steinerMapMutex.unlock();
                //check also if improves veryBest
                if (best.value < veryBest.value)
                    veryBest.value = best.value;
            }

        }				//end for i
        ///record overall best value for TpSet,
        //putting it with source -1.
        //this also serves as a proof that TpSet has been computed.
        steinerMapMutex.lock();
        X[Key(TpSet, -1)] = veryBest;
        steinerMapMutex.unlock();

        //printing
        if (veryBest.value < INF) {
            if (debugSteiner) cout << "Feasible: " << veryBest.value << endl;
        }
        else {
            if (debugSteiner) cout << "Infeasible." << endl;
        }
    } while (nextSubset(Tp, T.size(), maxTargetSubset));	//enumerate Tp subsets of T

    //printing
    if (g_verbose > 1 || debugSteiner) cout << "------\t Steiner update complete. Final size  : " << X.size();
    if (debugSteiner) cout << "  new sets = " << countSets;
    if (debugSteiner) cout << "  reachable / keys tested = " << countReachable << " / " << countKeys;
    if (g_verbose > 1 || debugSteiner) cout << endl;
}


/* compute the set of paths needed to reach 'cover' from 'source' according to the steiner map X.
 * paths are added in the set 'paths'
 */

void fillSolutionPaths(set<PathPair>& paths, int& source, Vertex_Set& cover, Steiner& X) {
    
    //look-up in the steiner map
    steinerMapMutex.lock();
    SteinerValue& sv = X[Key(cover, source)];
    steinerMapMutex.unlock();
    //if this belongs to a solution, there must be a finite value
    assert(sv.value < INF);
    //also, the tree can only have value 0 when the cover is a singleton equal to the source
    assert(sv.value > 0 || (cover.size() == 1 && cover.count(source) == 1));
    //base case: the tree is a simple path
    if (sv.singlePath) {
        //get the target
        int singleTarget = *cover.begin();
        //add the path
        paths.insert(pair<int, int>(source, singleTarget));
    }
    else {
        //general case: the tree branches in x= sv.steinerVertex
        int x = sv.steinerVertex;
        //add the path to x
        paths.insert(pair<int, int>(source, x));
        //add the paths of the subtrees
        fillSolutionPaths(paths, x, sv.leftPartition, X);
        fillSolutionPaths(paths, x, sv.rightPartition, X);
    }
}

/**** HyperSteinerForests ****/

/*An HyperSteiner has been computed, extract the set of trees needed to cover a given T
*/
void fillSteinerForest(HyperSteinerForest& F, HyperSteiner& Y, set<int>& T, Steiner& X) {

    HyperSteinerValue& y = Y[T];

    if (g_verbose > 2) {
        cout << "fillSteinerForest [ ";
        printVertexSet(T);
        cout << "]";
    }

    if (y.singleTree) {
        //T can be covered with a single tree, add it to the solution
        if (g_verbose > 2)  cout << " singleTree [v=" << y.value << "]\n";
        F.value += y.value;
        F.steinerTrees.push_back(HyperSteinerElement(y.source, T, y.rank));
        //add the corresponding paths
        fillSolutionPaths(F.solutionPaths, y.source, T, X);
    }
    else {
        //T needs to be covered by at least 2 different trees:
        if (g_verbose > 2) {
            cout << " Pair [";
            printVertexSet(y.leftPartition);
            cout << "  |  ";
            printVertexSet(y.rightPartition);
            cout << "]\n";
        }
        //recursive calls to each parts:)
        fillSteinerForest(F, Y, y.leftPartition, X);
        fillSteinerForest(F, Y, y.rightPartition, X);
    }
}

//self explicit
void  printHyperSteinerForest(HyperSteinerForest F) {
    if (F.value == INF) {
        std::cout << "[Forest] (no solution)" << endl;
    }
    else {
        std::cout << "[Forest] v=" << F.value;
        for (HyperSteinerElement e : F.steinerTrees) {
            std::cout << "\t Tree with source " << e.source << "\t leaves (rank=" << e.rank << ") =\t ";
            printVertexSet(e.cover);
            std::cout << endl;
        }
    }
}



/*print out the full set of (hyper-)edges needed in a solution */
void getFullSolution(HyperSteinerForest F, string fileName, const writeSolution & infoSol)
{
    if (F.value == INF) { std::cout << "[Forest] (no solution)" << endl; }
    else {
        string output = "";
            // Writting the hyperedge:
            for (auto idH : F.hyperId)
            {
                HyperArc myHyper =  infoSol.allHyperA[idH];
                auto subs = myHyper.substrates;
                for (size_t i = 0; i < subs.size(); ++i)
                {
                    if(i!=0)    {                        output += ",";                    }
                    output +=infoSol.name[subs[i]] ;
                }
                output += "\t";

                auto prods = myHyper.products;
                for (size_t i = 0; i < prods.size(); ++i)
                {
                    if(i!=0){                       output += ",";                    }
                    output += infoSol.name[prods[i]] ;
                }
                output += "\n";
            }
            // Writting the edges
            for (PathPair pathI : F.solutionPaths) {
                vector<vertex_t> thePath = getShortestPath(infoSol.name, infoSol.allPredecessors, pathI.first, pathI.second, false);
                string old = infoSol.name[*thePath.rbegin()];
                for (auto it = thePath.rbegin() + 1; it != thePath.rend(); ++it) {
                    output +=  old + "\t" + infoSol.name[*it] + "\n";
                    old = infoSol.name[*it];
                }
            }
            ofstream fichier(fileName, ios::out);  //writing solution
            if (fichier)  //
                   {
            fichier <<  output;
            fichier.close();
            if (g_verbose >1 ) {cout << "\n" << output;}

        }
        else
        {
            cerr << "Impossible d'ouvrir le fichier " << fileName << "!" << endl;
        }
    }
}

Weight minimum(Weight a, Weight b) {
    return (a < b ? a : b);
}

//function called for every thread.
/*!
 * \brief
 * \param borderVertices
 * \param S
 * \param T
 * \param nVertex
 * \param hyperArcs
 * `param DistanceMatrix
 * \param H
 * \param step :  current step, i.e: current "id" of the set of hyperarcs being used
 * \param infoSol : contains info about the predecessors, all hyperarcs and metabolites names, will allow us to print a solution
 */
void runOnPermutationOfHyperArcs(
   // vector<int> FullT, vector<int> FullS, set<int> FullTSet, Weight hyperArcsWeight, 
    const set<int>& borderVertices, const vector<int>& S, const vector<int>& T, 
    const int nVertex, const vector<HyperArc>& hyperArcs, const DistanceMatrix& sp,
    vector<int> H, const long double step, const writeSolution & myInfoSol, Steiner &threadX
    )
{

    const int verbose = g_verbose;
    //smallest value of H
    int min = 0;
    if (H.size() > 0) {
        min = H[0];
    }
    //build FullT and FullS and compute weight of H alone
    Weight hyperArcsWeight = 0;
    vector<int> FullT = T;
    vector<int> FullS = S;
    for (int h : H) {
        FullT.insert(FullT.end(), hyperArcs[h].substrates.begin(), hyperArcs[h].substrates.end());
        FullS.insert(FullS.end(), hyperArcs[h].products.begin(), hyperArcs[h].products.end());
        hyperArcsWeight += hyperArcs[h].weight;
    }


    if (hyperArcsWeight >= bestForest.value) {
        if (verbose > 0) std::cout << "  Expensive set of arcs: skip" << endl;
        //  countSteps -= boost::math::factorial<long double>(H.size());
        return;
    }

    


    //compute set equivalent of FullT, and remove duplicates in array
    Vertex_Set FullTSet(FullT.begin(), FullT.end());;
    FullT.assign(FullTSet.begin(), FullTSet.end());
    int FullTLength = FullT.size();
    if (verbose > 2) std::cout << "  FullTLength = " << FullTLength << endl;

    double substep = boost::math::factorial<long double>(H.size());

    do {//iterate over all permutations of H
        substep--;


        ////////////////////

    if (verbose > 2) {
        std::cout << "{" << step <<"."<<substep << "}\t  next permutation" << endl;
        std::cout << "{" << step << "." << substep << "}\t  H = [ ";
        for (auto i : H) std::cout << i << ' ';
        std::cout << "]  S = [ ";
        for (auto i : FullS) std::cout << i << ' ';
        std::cout << "]  T = [ ";
        for (auto i : FullT) std::cout << i << ' ';
        std::cout << "]  Weight =  " << hyperArcsWeight << endl;
    }


    //rank of original targets: bigger than all possible hyperarc rank
    const int targetTRank = hyperArcs.size() + 1;

    //map: to each border vertex give rank as source and as target (-1 if does not apply)
    map<int, int> sRank;
    map<int, int> tRank;

    //if the arcs themselves are heavier than the best solution: no hope whatsoever.
    if (hyperArcsWeight >= bestForest.value) {
        //    countSteps+= boost::math::factorial<long double>(HLength); 

        return; //continue;
    }

    //HyperSteiner= set of steiner trees satisfying ranks of hyperarcs
    HyperSteiner Y;

    /* compute rank for each vertex:
     *  -1 = not a source/target here
     *  0  (in sRank) = an original source
     *  r  = a product (sRank) or a substrate (tRank) of hyperarc H[r], not appearing before
     *  targetTRank = an original target, not source of any hyperarc.
    */
    for (int i : borderVertices) {
        sRank[i] = -1;
        tRank[i] = -1;
    }
    for (int i : S) {
        sRank[i] = 0;
    }
    for (int i : T) {
        tRank[i] = targetTRank;
    }
    int rank = 1;
    //in general, the rank is the index of the first hyperarc in which a vertex appears
    for (int h : H) {
        for (int i : hyperArcs[h].products) {
            if (sRank[i] == -1) sRank[i] = rank;
        }
        for (int i : hyperArcs[h].substrates) {
            if (tRank[i] == -1 || tRank[i] == targetTRank) tRank[i] = rank;
        }
        rank++;
    }

    //print
    if (verbose >= 4) {
        for (const auto &p : sRank) {
            std::cout << "{" << step << "." << substep <<"}\t   sRank[" << p.first << "] = " << p.second << '\n';
        }
        for (const auto &p : tRank) {
            std::cout << "{" << step << "." << substep << "}\t  tRank[" << p.first << "] = " << p.second << '\n';
        }
    }

    //check that all targets are reachable 
    //(i.e. connected to a source of lower rank)
    bool allReachable = true;
    Weight targetLowerBound = INF;
    Weight sourceLowerBound = INF;
    for (int t : FullT) {
        bool reachable = false;
        for (int s : FullS) {
            if (sRank[s] < tRank[t] && SP(s, t) < INF) {
                reachable = true;
                targetLowerBound = minimum(targetLowerBound, SP(s, t));
                if (sRank[s] == 0) {
                    sourceLowerBound = minimum(sourceLowerBound, SP(s, t));
                }
            }
        }
        if (!reachable) {
            allReachable = false;
            if (verbose > 1) {
                std::cout << "{" << step << "." << substep << "}\t non-reachable element: " << t << " from ";
                print(S);
                std::cout<< endl;
            }
            break;
        }
    }
    if (!allReachable) continue;
    targetLowerBound += hyperArcsWeight;
    

    //check that all hyperarcs are possibly useful
    // (i.e., each hyperarc has at least 1 product connected to a target of higher rank)
    bool allUseful = true;
    for (int h : H) { 
        bool useful = false;
        Weight hLowerBound = INF;
        for (int s : hyperArcs[h].products) {
            for (int t : FullT) {
                if (sRank[s] < tRank[t] && SP(s, t) < INF) {
                    useful = true;
                    hLowerBound = minimum(hLowerBound, SP(s, t));
                }
            }
        }
        sourceLowerBound += hLowerBound;
        if (!useful) {
            if (verbose > 1) {
                printMutex.lock();
                std::cout << "{" << step << "." << substep << "}\t useless hyperarc: " << h <<endl;
                printMutex.unlock();
            }
            allUseful = false;
        }          
    }
    sourceLowerBound += hyperArcsWeight;
    if (verbose > 0) {
        printMutex.lock();
        std::cout << "{" << step << "." << substep << "}\t feasible,";
        if (allUseful) {
            std::cout << " and each one is useful;";
            std::cout << " lower bounds computed: (T) " << targetLowerBound << " (S) " << sourceLowerBound;
        }
        else
            std::cout << " but one is useless;                                          ";
        std::cout<<" \t H = ";
        print(H);
        printMutex.unlock();
    }
    if (!allUseful) continue;
    
    if (g_use_lowerbounds) {
        bestSolutionMutex.lock();
        Weight opt = bestForest.value;
        bestSolutionMutex.unlock();
        if (targetLowerBound >= opt || sourceLowerBound >= opt) {
            printMutex.lock();
            std::cout<<"{" << step << "." << substep << "}\t skipped based on lower bounds.\n";
            printMutex.unlock();
            continue;
        }

    }

    
    if (g_dry_run) continue;


    //compute optimal steiner trees to reach all subsets targets from any vertex


    if (!steinerHasBeenDone(threadX, FullTSet)) {
        updateSteiner(threadX, FullT, nVertex, -1, sp);
        if (verbose > 2) std::cout << "  Steiner updated." << endl;
    }
    else {
        if (verbose > 2) std::cout << "  Steiner already computed." << endl;
    }



    //successive subsets of FullT
    vector<int> Tp;
    Tp.push_back(0);


    do {//enumerate Tp
        const int TpLength = Tp.size();

        //placeholder for best HyperSteiner covering Tp
        HyperSteinerValue best;
        best.value = INF;

        //convert Tp to set
        Vertex_Set TpSet = getSubset(FullT, Tp);
        checkKeyIsPresent(threadX, TpSet, "hypersteiner-1");

        //print
        if (verbose > 3) std::cout << "{" << step << "." << substep << "}\t   next Tp: ";
        if (verbose > 3) printVertexSet(TpSet);
        if (verbose > 3) std::cout << ";";
        if (verbose == 2) {
            std::cout << ".";
        }

        //compute rank of Tp (lowest hyperarc of which some element of Tp is a head)
        int rank = targetTRank;
        for (int t : Tp) {
            if (tRank[FullT[t]] < rank) {
                rank = tRank[FullT[t]];
            }
        }
        if (verbose > 3)  std::cout << " rank=" << rank;

        //look at possible roots of steiner trees
        for (int i : borderVertices) {
            //only those whose rank is valid
            if (sRank[i] != -1 && sRank[i] < rank) {
                //try to directly cover all Tp with only i as root

                checkKeyIsPresent(threadX, TpSet, "hypersteiner-2");
                Weight v = getSteinerValue(threadX, TpSet, i);
                assert(v >= 0);
                //record if good value
                if (v < best.value) {
                    best.value = v;
                    best.singleTree = true;
                    best.source = i;
                    best.rank = rank;
                }
            }
        }
        if (verbose > 3)  std::cout << " base value=" << best.value;
        //now try to split Tp in 2 (to have different roots)
        if (TpLength > 1) {

            //successive subsets of Tp
            vector<int> Tpa;
            vector<int> Tpb;
            //start with a singleton
            Tpa.push_back(0);


            do {//enumerate Tpa subset of Tp            

                //Compute Tpb = Tp\Tpa
                Tpb = complement(Tpa, TpLength);

                //Get Tpa and Tpb in "vertex set" format
                Vertex_Set TpaSet = getSubsubset(FullT, Tp, Tpa);
                Vertex_Set TpbSet = getSubsubset(FullT, Tp, Tpb);

                //fetch best steiner forests for Tpa and Tpb
                Weight ya = Y[TpaSet].value;
                Weight yb = Y[TpbSet].value;

                if (verbose > 3) std::cout << ".";

                //check that both subforests are feasible
                if (ya == INF || yb == INF) continue; //next partition

                if (verbose > 3) std::cout << " fine " << ya << " " << yb;

                //value of this partition
                Weight y = ya + yb;
                assert(y >= 0);
                //record if good value
                if (y < best.value) {
                    best.value = y;
                    best.singleTree = false;
                    best.leftPartition = TpaSet;
                    best.rightPartition = TpbSet;
                }

            } while (nextSubset(Tpa, TpLength, (int)TpLength / 2));        //enumerate Tpa
        }//end if |Tp|>1

        //now all options for Tp have been tested: record
        if (verbose > 3) std::cout << " done = " << best.value << "\n";
        assert(best.value >= 0);
        Y[TpSet] = best;
    } while (nextSubset(Tp, FullT.size(), FullT.size()));        //enumerate Tp
    if (verbose == 2)  std::cout << endl;

    //We're only interested in the forest covering all FullT
    const HyperSteinerValue& y = Y[FullTSet];

    if (y.value == INF) {
        if (verbose > 0) std::cout << "{" << step << "." << substep << "}\t actually has no solution, some strange case happened." << endl;
    }
    else {

        //record if good value
        HyperSteinerForest forest;
        if (verbose > 0) std::cout << "{" << step << "." << substep << "}\t Here is the solution :  " << endl;
        forest.value = hyperArcsWeight;
        forest.hyperId = H;

        fillSteinerForest(forest, Y, FullTSet, threadX);
		string solF = outDir + "/solution_" + std::to_string(step) +  "." + std::to_string(substep) + ".txt";
   		cout << "writing in " << solF << endl;
   		getFullSolution(forest, solF, myInfoSol);
        if (verbose > 0) {
            std::cout << "{" << step << "." << substep << "}\t ** With   H = [ ";
            for (auto i : H) std::cout << i << ' ';
            std::cout << " ]   " << endl;
            printHyperSteinerForest(forest);
        }


        bestSolutionMutex.lock();
        if (bestForest.value > forest.value) {
            bestForest = forest;
            cout << "{" << step << "." << substep << "}\t**** This is the current best solution ****\n";
        }
        else {
            cout << "{" << step << "." << substep << "}\t Some earlier solution was better\n";
        }
        bestSolutionMutex.unlock();


    }

    // filtering out dummy solutions?
    // - with a reversible reaction used in both directions
    // - with a useless hyperarc (=no path starts from any of its products?)

    


        } while (nextPermutation(H, min));  // enumerate permutations of H 

        if (verbose > 1 && (fmod(step, g_step_print_progress) == 0)) {
            printMutex.lock();
            std::cout << "{" << step << "-done}\t" << endl;
            printMutex.unlock();
        }     
}


void runOnFractionOfInstances(
    // vector<int> FullT, vector<int> FullS, set<int> FullTSet, Weight hyperArcsWeight, 
    const set<int>& borderVertices, const vector<int>& S, const vector<int>& T,
    const int nVertex, const vector<HyperArc>& hyperArcs, const DistanceMatrix& sp,
    vector<int> firstH, const int threadId, const int firstStep, const writeSolution & myInfoSol
    )
{
    const int verbose = g_verbose;
    int countSteps = firstStep;
    vector<int> H = firstH;
    if (verbose > 0) {
        printMutex.lock();
        std::cout << "Thread "<< threadId << " starts with first H=";
        print(firstH);
        printMutex.unlock();
    }
    Steiner X;
    bool skipThis = false;
    do {//enumerate H subset of {0, ... ,   hyperArcs.size()-1}     


        if (verbose > 1) {
            printMutex.lock();
            std::cout << "{" << countSteps << "-launching in thread " << threadId << "} with H=";
            print(H);
            printMutex.unlock();
        }
        
        if (g_reset_from_earlier) {
            if (countSteps <= g_reset_from.at(threadId) ) {
                if (skipThis && verbose > 0) {
                    printMutex.lock();
                    std::cout << "{" << countSteps << "}  Thread " << threadId << " Starting here." << endl;
                    printMutex.unlock();
                    skipThis = false;
                }
            }
            else {
                if (verbose > 1) {
                    printMutex.lock();
                    std::cout << "{" << countSteps << " Thread " << threadId << "} skipped";
                    printMutex.unlock();
                }
                skipThis = true;
            }
        }
        if (!skipThis) {
            runOnPermutationOfHyperArcs(
                borderVertices, S, T, nVertex, hyperArcs, sp, H, countSteps, myInfoSol, X
                );
        }

        countSteps -= g_nb_threads;
        
        globalCountMutex.lock();
        threadProgress[threadId] = countSteps;
        globalCount--;
        if (verbose > 0 && fmod(globalCount, g_step_print_progress) == 0) {
            printMutex.lock();
            std::cout << "Steps remaining : \t" << globalCount << ". Thread detail: ";
            print(threadProgress);
            printMutex.unlock();
        }
        globalCountMutex.unlock();

    } while (nextKSubset(H, hyperArcs.size(), g_max_tentacules, g_nb_threads)); //enumerate subsets of H
    printMutex.lock();
    printSteinerStats(X, threadId);
    printMutex.unlock();
}


HyperSteinerForest parallelHyperSteiner(vector<int>& S, vector<int>& T, int nVertex, vector<HyperArc>& hyperArcs, DistanceMatrix& sp, const writeSolution & infoSol)
{
    //testThread();
     //verbose level: 1..5
    const int verbose = g_verbose;
    if (verbose > 0)    std::cout << "max_tentacules : " << g_max_tentacules<< " (total="<< hyperArcs.size()<<")" << endl;

    //initialize best solution
    bestForest.value = INF;

    if (g_reset_from_earlier) {
        std::cout << "Warning,  RESET_FROM_EARLIER=true, the beginning of the search space will be skipped." << endl;
    }

    //thread children
    std::queue<std::thread> allthreads;

    // build border vertices array :
    // all sources and targets of any hyperarc, with S and T
    set<int> borderVertices;
    if (verbose > 0) std::cout << nVertex << " vertices\t ";
    if (verbose > 0) std::cout << "HyperSteiner   S=";
    for (int i : S) {
        if (verbose > 0) std::cout << i;
        borderVertices.insert(i);
    }
    if (verbose > 0) std::cout << "   T= ";
    for (int i : T) {
        if (verbose > 0) std::cout << i;
        borderVertices.insert(i);
    }
    for (HyperArc h : hyperArcs) {
        for (int i : h.substrates) {
            borderVertices.insert(i);
        }
        for (int i : h.products) {
            borderVertices.insert(i);
        }
    }
    if (verbose > 0) std::cout << "\n";
    
    steinerNodes = enumerateSteinerNodes(S,T,nVertex, hyperArcs,sp);

    //count total number of steps (permutations of subsets of hyperarcs)
    long double  totalSteps = 1;
    long double  countSteps = 0;
    long double nsubsets = 1;
    for (int i = 1; i<= g_max_tentacules; i ++) {
        //totalSteps = 1 + (hyperArcs.size() - i)*totalSteps;
        nsubsets = (nsubsets * (hyperArcs.size() - i + 1)) / i;
        totalSteps += nsubsets;
        //std::cout << (hyperArcs.size() - i + 1) <<" "<< nsubsets << "\n";
    }
    countSteps = totalSteps;
    globalCount = totalSteps;
    std::cout << "Total number of steps " << globalCount<<"\n";

    vector<int> H;

    time_t start = time(0);
    for (int i = 0; i < g_nb_threads; i++) {
        //enumerate H subset of {0, ... ,   hyperArcs.size()-1}     
        countSteps--;
        if (verbose > 1) std::cout << "Launching Thread "<<i << endl;
        
        globalCountMutex.lock();
        threadProgress.push_back(countSteps);
        globalCountMutex.unlock();
        
        allthreads.push(std::thread(
                runOnFractionOfInstances, 
                std::ref(borderVertices), std::ref(S), std::ref(T), nVertex, 
             std::ref(hyperArcs), std::ref(sp), H, 
             i, countSteps, std::ref(infoSol)
                ));
         if (!nextSubset(H, hyperArcs.size(), g_max_tentacules)) {
             break;
         }
    }  //enumerate subsets of H


    //all subsets are done: wait for the last threads to finish
    std::cout << "{x}\t waiting" << endl;
    while (!allthreads.empty()) {
        allthreads.front().join();
        allthreads.pop();
    }
    std::cout << "{x}\t complete!" << endl;
    time_t end = time(0);
    std::cout << "time : " << difftime(end,start) << "s\n";
    if (g_keep_stats) {
        std::cout << "STATS: \t Steiner Trees computed: " << steinerKeys << " \t\t Steiner Trees Reused: " << steinerReused << endl;
    }
   /// printSteinerStats(globalX);
   // if (verbose > 2)  printQuickSteiner(globalX);
    return bestForest;
}



//test function
int test()
{
    return 0;
}
//thread testing


void func(int tid) {
    std::cout << "Launched by thread " << tid << std::endl;
}

int testThread() {
    std::vector<std::thread> th;
    int nr_threads = 10;
    //Launch a group of threads
    for (int i = 0; i < nr_threads; ++i) {
        th.push_back(std::thread(func, i));
    }
    //Join the threads with the main thread
    for (auto &t : th) {
        t.join();
    }
    return 0;
}
